import asyncio
import base64
from collections import OrderedDict
from contextlib import closing
from datetime import timedelta
import enum
import inspect
from itertools import chain
import numbers
from pathlib import Path
import random
import re
import sys
import socket
from typing import (
    Any, Union, Type,
    Iterator, Callable, Awaitable,
    Tuple,
    cast,
)
import uuid

import aiohttp
from async_timeout import timeout as _timeout
import janus

from .types import BinarySize, Sentinel


def env_info():
    """
    Returns a string that contains the Python version and runtime path.
    """
    v = sys.version_info
    pyver = f'Python {v.major}.{v.minor}.{v.micro}'
    if v.releaselevel == 'alpha':
        pyver += 'a'
    if v.releaselevel == 'beta':
        pyver += 'b'
    if v.releaselevel == 'candidate':
        pyver += 'rc'
    if v.releaselevel != 'final':
        pyver += str(v.serial)
    return f'{pyver} (env: {sys.prefix})'


def odict(*args):
    """
    A short-hand for the constructor of OrderedDict.
    :code:`odict(('a',1), ('b',2))` is equivalent to
    :code:`OrderedDict([('a',1), ('b',2)])`.
    """
    return OrderedDict(args)


def dict2kvlist(o):
    """
    Serializes a dict-like object into a generator of the flatten list of
    repeating key-value pairs.  It is useful when using HMSET method in Redis.

    Example:
    >>> list(dict2kvlist({'a': 1, 'b': 2}))
    ['a', 1, 'b', 2]
    """
    return chain.from_iterable((k, v) for k, v in o.items())


def generate_uuid() -> str:
    u = uuid.uuid4()
    # Strip the last two padding characters because u always has fixed length.
    return base64.urlsafe_b64encode(u.bytes)[:-2].decode('ascii')


def get_random_seq(length: float, num_points: int, min_distance: float) -> Iterator[float]:
    """
    Generate a random sequence of numbers within the range [0, length]
    with the given number of points and the minimum distance between the points.

    Note that X ( = the minimum distance d x the number of points N) must be equivalent to or smaller than
    the length L + d to guarantee the the minimum distance between the points.
    If X == L + d, the points are always equally spaced with d.

    :return: An iterator over the generated sequence
    """
    assert num_points * min_distance <= length + min_distance, \
           'There are too many points or it has a too large distance which cannot be fit into the given length.'
    extra = length - (num_points - 1) * min_distance
    ro = [random.uniform(0, 1) for _ in range(num_points + 1)]
    sum_ro = sum(ro)
    rn = [extra * r / sum_ro for r in ro[0:num_points]]
    spacing = [min_distance + rn[i] for i in range(num_points)]
    cumulative_sum = 0.0
    for s in spacing:
        cumulative_sum += s
        yield cumulative_sum - min_distance


def find_free_port(bind_addr: str = '127.0.0.1') -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind((bind_addr, 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def nmget(o, key_path, def_val=None, path_delimiter='.', null_as_default=True):
    """
    A short-hand for retrieving a value from nested mappings
    ("nested-mapping-get"). At each level it checks if the given "path"
    component in the given key exists and return the default value whenever
    fails.

    Example:
    >>> o = {'a':{'b':1}, 'x': None}
    >>> nmget(o, 'a', 0)
    {'b': 1}
    >>> nmget(o, 'a.b', 0)
    1
    >>> nmget(o, 'a/b', 0, '/')
    1
    >>> nmget(o, 'a.c', 0)
    0
    >>> nmget(o, 'x', 0)
    0
    >>> nmget(o, 'x', 0, null_as_default=False)
    None
    """
    pieces = key_path.split(path_delimiter)
    while pieces:
        p = pieces.pop(0)
        if o is None or p not in o:
            return def_val
        o = o[p]
    if o is None and null_as_default:
        return def_val
    return o


def readable_size_to_bytes(expr):
    if isinstance(expr, numbers.Real):
        return BinarySize(expr)
    return BinarySize.from_str(expr)


def str_to_timedelta(tstr):
    """
    Convert humanized timedelta string into a Python timedelta object.

    Example:
    >>> str_to_timedelta('30min')
    datetime.timedelta(seconds=1800)
    >>> str_to_timedelta('1d1hr')
    datetime.timedelta(days=1, seconds=3600)
    >>> str_to_timedelta('2hours 15min')
    datetime.timedelta(seconds=8100)
    >>> str_to_timedelta('20sec')
    datetime.timedelta(seconds=20)
    >>> str_to_timedelta('300')
    datetime.timedelta(seconds=300)
    >>> str_to_timedelta('-1day')
    datetime.timedelta(days=-1)
    """
    _rx = re.compile(r'(?P<sign>[+|-])?\s*'
                     r'((?P<days>\d+(\.\d+)?)(d|day|days))?\s*'
                     r'((?P<hours>\d+(\.\d+)?)(h|hr|hrs|hour|hours))?\s*'
                     r'((?P<minutes>\d+(\.\d+)?)(m|min|mins|minute|minutes))?\s*'
                     r'((?P<seconds>\d+(\.\d+)?)(s|sec|secs|second|seconds))?$')
    ts = _rx.match(tstr)
    if not ts:
        try:
            return timedelta(seconds=float(tstr))  # consider bare number string as seconds
        except TypeError:
            pass
        raise ValueError('Invalid time expression')
    ts = ts.groupdict()
    sign = ts.pop('sign', None)
    if set(ts.values()) == {None}:
        raise ValueError('Invalid time expression')
    params = {n: -float(t) if sign == '-' else float(t) for n, t in ts.items() if t}
    return timedelta(**params)


async def curl(url, default_value=None, params=None, headers=None, timeout=0.2):
    try:
        async with aiohttp.ClientSession() as sess:
            with _timeout(timeout):
                async with sess.get(url, params=params, headers=headers) as resp:
                    assert resp.status == 200
                    body = await resp.text()
                    return body.strip()
    except (asyncio.TimeoutError, aiohttp.ClientError, AssertionError):
        if callable(default_value):
            return default_value()
        return default_value


class StringSetFlag(enum.Flag):

    def __eq__(self, other):
        return self.value == other

    def __hash__(self):
        return hash(self.value)

    def __or__(self, other):
        if isinstance(other, type(self)):
            other = other.value
        if not isinstance(other, (set, frozenset)):
            other = set((other,))
        return set((self.value,)) | other

    __ror__ = __or__

    def __and__(self, other):
        if isinstance(other, (set, frozenset)):
            return self.value in other
        if isinstance(other, str):
            return self.value == other
        raise TypeError

    def __rand__(self, other):
        if isinstance(other, (set, frozenset)):
            return self.value in other
        if isinstance(other, str):
            return self.value == other
        raise TypeError

    def __xor__(self, other):
        if isinstance(other, (set, frozenset)):
            return set((self.value,)) ^ other
        if isinstance(other, str):
            if other == self.value:
                return set()
            else:
                return other
        raise TypeError

    def __rxor__(self, other):
        if isinstance(other, (set, frozenset)):
            return other ^ set((self.value,))
        if isinstance(other, str):
            if other == self.value:
                return set()
            else:
                return other
        raise TypeError

    def __str__(self):
        return self.value


class AsyncBarrier:
    """
    This class provides a simplified asyncio-version of threading.Barrier class.
    """

    num_parties: int = 1
    cond: asyncio.Condition

    def __init__(self, num_parties: int):
        self.num_parties = num_parties
        self.count = 0
        self.cond = asyncio.Condition()

    async def wait(self):
        async with self.cond:
            self.count += 1
            if self.count == self.num_parties:
                self.cond.notify_all()
            else:
                while self.count < self.num_parties:
                    await self.cond.wait()

    def reset(self):
        self.count = 0
        # FIXME: if there are waiting coroutines, let them
        #        raise BrokenBarrierError like threading.Barrier


class FstabEntry:
    """
    Entry class represents a non-comment line on the `fstab` file.
    """
    def __init__(self, device, mountpoint, fstype, options, d=0, p=0):
        self.device = device
        self.mountpoint = mountpoint
        self.fstype = fstype
        if not options:
            options = 'defaults'
        self.options = options
        self.d = d
        self.p = p

    def __eq__(self, o):
        return str(self) == str(o)

    def __str__(self):
        return "{} {} {} {} {} {}".format(self.device,
                                          self.mountpoint,
                                          self.fstype,
                                          self.options,
                                          self.d,
                                          self.p)


class Fstab:
    """
    Reader/writer for fstab file.
    Takes aiofile pointer for async I/O. It should be writable if add/remove
    operations are needed.

    NOTE: This class references Jorge Niedbalski R.'s gist snippet.
          We have been converted it to be compatible with Python 3
          and to support async I/O.
          (https://gist.github.com/niedbalski/507e974ed2d54a87ad37)
    """
    def __init__(self, fp):
        self._fp = fp

    def _hydrate_entry(self, line):
        return FstabEntry(*[x for x in line.strip('\n').split(' ') if x not in ('', None)])

    async def get_entries(self):
        await self._fp.seek(0)
        for line in await self._fp.readlines():
            try:
                line = line.strip()
                if not line.startswith("#"):
                    yield self._hydrate_entry(line)
            except TypeError:
                pass

    async def get_entry_by_attr(self, attr, value):
        async for entry in self.get_entries():
            e_attr = getattr(entry, attr)
            if e_attr == value:
                return entry
        return None

    async def add_entry(self, entry):
        if await self.get_entry_by_attr('device', entry.device):
            return False
        await self._fp.write(str(entry) + '\n')
        await self._fp.truncate()
        return entry

    async def add(self, device, mountpoint, fstype, options=None, d=0, p=0):
        return await self.add_entry(FstabEntry(device, mountpoint, fstype, options, d, p))

    async def remove_entry(self, entry):
        await self._fp.seek(0)
        lines = await self._fp.readlines()
        found = False
        for index, line in enumerate(lines):
            try:
                if not line.strip().startswith("#"):
                    if self._hydrate_entry(line) == entry:
                        found = True
                        break
            except TypeError:
                pass
        if not found:
            return False
        lines.remove(line)
        await self._fp.seek(0)
        await self._fp.write(''.join(lines))
        await self._fp.truncate()
        return True

    async def remove_by_mountpoint(self, mountpoint):
        entry = await self.get_entry_by_attr('mountpoint', mountpoint)
        if entry:
            return await self.remove_entry(entry)
        return False


current_loop: Callable[[], asyncio.AbstractEventLoop]
if hasattr(asyncio, 'get_running_loop'):
    current_loop = asyncio.get_running_loop  # type: ignore
else:
    current_loop = asyncio.get_event_loop    # type: ignore


async def run_through(
    *awaitable_or_callables: Union[Callable[[], None], Awaitable[None]],
    ignored_exceptions: Tuple[Type[Exception], ...],
) -> None:
    """
    A syntactic sugar to simplify the code patterns like:

    .. code-block:: python3

       try:
           await do1()
       except MyError:
           pass
       try:
           await do2()
       except MyError:
           pass
       try:
           await do3()
       except MyError:
           pass

    Using ``run_through()``, it becomes:

    .. code-block:: python3

       await run_through(
           do1(),
           do2(),
           do3(),
           ignored_exceptions=(MyError,),
       )
    """
    for f in awaitable_or_callables:
        try:
            if inspect.iscoroutinefunction(f):
                await f()  # type: ignore
            elif inspect.isawaitable(f):
                await f  # type: ignore
            else:
                f()  # type: ignore
        except Exception as e:
            if isinstance(e, cast(Tuple[Any, ...], ignored_exceptions)):
                continue
            raise


class AsyncFileWriter:
    """
    This class provides a context manager for making sequential async
    writes using janus queue.
    """
    def __init__(
            self,
            loop: asyncio.AbstractEventLoop,
            target_filename: Union[str, Path],
            access_mode: str,
            decode: Callable[[str], bytes] = None,
            max_chunks: int = None) -> None:
        if max_chunks is None:
            max_chunks = 0
        self._q: janus.Queue[Union[bytes, str, Sentinel]] = janus.Queue(maxsize=max_chunks)
        self._loop = loop
        self._target_filename = target_filename
        self._access_mode = access_mode
        self._decode = decode

    async def __aenter__(self):
        self._fut = self._loop.run_in_executor(None, self._write)
        return self

    def _write(self):
        with open(self._target_filename, self._access_mode) as f:
            while True:
                item = self._q.sync_q.get()
                if item is Sentinel.TOKEN:
                    break
                if self._decode is not None:
                    item = self._decode(item)
                f.write(item)
                self._q.sync_q.task_done()

    async def __aexit__(self, exc_type, exc, tb):
        await self._q.async_q.put(Sentinel.TOKEN)
        try:
            await self._fut
        finally:
            self._q.close()
            await self._q.wait_closed()

    async def write(self, item):
        await self._q.async_q.put(item)



from unittest import mock
try:
    # Since Python 3.8, AsyncMock is now part of the stdlib.
    # Python 3.8 also adds magic-mocking async iterators and async context managers.
    from unittest.mock import AsyncMock  # type: ignore
except ImportError:
    from asynctest import CoroutineMock as AsyncMock  # type: ignore


def mock_corofunc(return_value):
    """
    Return mock coroutine function.

    Python's default mock module does not support coroutines.
    """
    async def _mock_corofunc(*args, **kargs):
        return return_value
    return mock.Mock(wraps=_mock_corofunc)


async def mock_awaitable(**kwargs):
    """
    Mock awaitable.

    An awaitable can be a native coroutine object "returned from" a native
    coroutine function.
    """
    return AsyncMock(**kwargs)


class AsyncContextManagerMock:
    """
    Mock async context manager.

    Can be used to get around `async with` statement for testing.
    Must implement `__aenter__` and `__aexit__` which returns awaitable.
    Attributes of the awaitable (and self for convenience) can be set by
    passing `kwargs`.
    """
    def __init__(self, *args, **kwargs):
        self.context = kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

    async def __aenter__(self):
        return AsyncMock(**self.context)

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        pass



from __future__ import annotations

from abc import ABCMeta, abstractmethod
import enum
from typing import (
    Any,
    Mapping,
    Union,
)

from . import AbstractPlugin, BasePluginContext

__all__ = (
    'AbstractStatReporterPlugin',
    'AbstractErrorReporterPlugin',
    'StatsPluginContext',
    'ErrorPluginContext',
    'INCREMENT',
    'GAUGE',
)


class StatMetricTypes(enum.Enum):
    INCREMENT = 0
    GAUGE = 1


INCREMENT = StatMetricTypes.INCREMENT
GAUGE = StatMetricTypes.GAUGE


class AbstractStatReporterPlugin(AbstractPlugin, metaclass=ABCMeta):

    async def init(self, context: Any = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    @abstractmethod
    async def report_metric(
        self,
        metric_type: StatMetricTypes,
        metric_name: str,
        value: Union[float, int] = None,
    ) -> None:
        pass


class AbstractErrorReporterPlugin(AbstractPlugin, metaclass=ABCMeta):
    async def init(self, context: Any = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    @abstractmethod
    async def capture_exception(
        self,
        exc_instance: Exception = None,
        context: Mapping[str, Any] = None,
    ) -> None:
        pass

    @abstractmethod
    async def capture_message(self, message: str) -> None:
        pass


class StatsPluginContext(BasePluginContext[AbstractStatReporterPlugin]):
    plugin_group = 'backendai_stats_monitor_v20'

    async def report_metric(
        self,
        metric_type: StatMetricTypes,
        metric_name: str,
        value: Union[float, int] = None,
    ) -> None:
        for plugin_instance in self.plugins.values():
            await plugin_instance.report_metric(metric_type, metric_name, value)


class ErrorPluginContext(BasePluginContext[AbstractErrorReporterPlugin]):
    plugin_group = 'backendai_error_monitor_v20'

    async def capture_exception(
        self,
        exc_instance: Exception = None,
        context: Mapping[str, Any] = None,
    ) -> None:
        for plugin_instance in self.plugins.values():
            await plugin_instance.capture_exception(exc_instance, context)

    async def capture_message(self, message: str) -> None:
        for plugin_instance in self.plugins.values():
            await plugin_instance.capture_message(message)



import re
from typing import (
    List,
    Sequence,
    Set,
    Type,
)

from .types import ServicePort, ServicePortProtocols

__all__ = (
    'parse_service_ports'
)

_rx_service_ports = re.compile(
    r'^(?P<name>[\w-]+):(?P<proto>\w+):(?P<ports>\[\d+(?:,\d+)*\]|\d+)(?:,|$)')


def parse_service_ports(s: str, exception_cls: Type[Exception] = None) -> Sequence[ServicePort]:
    items: List[ServicePort] = []
    if exception_cls is None:
        exception_cls = ValueError
    used_ports: Set[int] = set()
    while True:
        match = _rx_service_ports.search(s)
        if match:
            s = s[len(match.group(0)):]
            name = match.group('name')
            if not name:
                raise exception_cls('Service port name must be not empty.')
            protocol = match.group('proto')
            if protocol == 'pty':
                # unsupported, skip
                continue
            if protocol not in ('tcp', 'http', 'preopen'):
                raise exception_cls(f'Unsupported service port protocol: {protocol}')
            ports = tuple(map(int, match.group('ports').strip('[]').split(',')))
            for p in ports:
                if p in used_ports:
                    raise exception_cls(f'The port {p} is already used by another service port.')
                if p <= 1024:
                    raise exception_cls(f'The service port number {p} must be '
                                        f'larger than 1024 to run without the root privilege.')
                if p >= 65535:
                    raise exception_cls(f'The service port number {p} must be smaller than 65535.')
                if p in (2000, 2001, 2002, 2003, 2200, 7681):
                    raise exception_cls('The service ports 2000 to 2003, 2200 and 7681 '
                                        'are reserved for internal use.')
                used_ports.add(p)
            items.append({
                'name': name,
                'protocol': ServicePortProtocols(protocol),
                'container_ports': ports,
                'host_ports': (None,) * len(ports),
            })
        else:
            if len(s) > 0:
                raise exception_cls('Invalid format')
            break
    return items



from __future__ import annotations

from abc import ABCMeta, abstractmethod
import asyncio
import logging
import pkg_resources
import re
from typing import (
    Any,
    ClassVar,
    Container,
    Dict,
    Generic,
    Iterator,
    Mapping,
    Set,
    Tuple,
    Type,
    TypeVar,
)

from ..etcd import AsyncEtcd
from ..types import QueueSentinel
from ..logging_utils import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__name__))

__all__ = (
    'AbstractPlugin',
    'BasePluginContext',
)


class AbstractPlugin(metaclass=ABCMeta):
    """
    The minimum generic plugin interface.
    """

    plugin_config: Mapping[str, Any]
    """
    ``plugin_config`` contains the plugin-specific configuration read from the etcd.
    """

    local_config: Mapping[str, Any]
    """
    ``local_config`` contains the configuration read from the disk TOML file of the current daemon.
    This configuration is only updated when restarting the daemon and thus plugins should assume
    that it's read-only and immutable during its lifetime.
    e.g., If the plugin is running with the manager, it's the validated content of manager.toml file.
    """

    config_watch_enabled: ClassVar[bool] = True
    """
    If set True (default), the hosting plugin context will watch and automatically update
    the etcd's plugin configuration changes via the ``update_plugin_config()`` method.
    """

    def __init__(self, plugin_config: Mapping[str, Any], local_config: Mapping[str, Any]) -> None:
        """
        Instantiate the plugin with the given initial configuration.
        """
        self.plugin_config = plugin_config
        self.local_config = local_config

    @abstractmethod
    async def init(self, context: Any = None) -> None:
        """
        Initialize any resource used by the plugin.
        """
        pass

    @abstractmethod
    async def cleanup(self) -> None:
        """
        Clean up any resource used by the plugin upon server cleanup.
        """
        pass

    @abstractmethod
    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        """
        Handle runtime configuration updates.
        The config parameter contains both the updated parts
        and unchanged parts of the configuration.

        The default implementation is just to replace the config property,
        but actual plugins may trigger other operations to reflect config changes
        and/or inspect the differences of configs before replacing the current config.
        """
        self.plugin_config = plugin_config


P = TypeVar('P', bound=AbstractPlugin)


class BasePluginContext(Generic[P]):
    """
    A minimal plugin manager which controls the lifecycles of the given plugins
    and watches & applies the configuration changes in etcd.

    The subclasses must redefine ``plugin_group``.
    """

    etcd: AsyncEtcd
    local_config: Mapping[str, Any]
    plugins: Dict[str, P]
    plugin_group: ClassVar[str] = 'backendai_XXX_v10'

    _config_watchers: Set[asyncio.Task]

    def __init__(self, etcd: AsyncEtcd, local_config: Mapping[str, Any]) -> None:
        self.etcd = etcd
        self.local_config = local_config
        self.plugins = {}
        self._config_watchers = set()
        if m := re.search(r'^backendai_(\w+)_v(\d+)$', self.plugin_group):
            self._group_key = m.group(1)
        else:
            raise TypeError(
                f"{type(self).__name__} has invalid plugin_group class attribute",
                self.plugin_group,
            )

    @classmethod
    def discover_plugins(
        cls,
        plugin_group: str,
        blocklist: Container[str] = None,
    ) -> Iterator[Tuple[str, Type[P]]]:
        if blocklist is None:
            blocklist = set()
        for entrypoint in pkg_resources.iter_entry_points(plugin_group):
            if entrypoint.name in blocklist:
                continue
            log.info('loading plugin (group:{}): {}', plugin_group, entrypoint.name)
            yield entrypoint.name, entrypoint.load()

    async def init(self, context: Any = None) -> None:
        scanned_plugins = self.discover_plugins(self.plugin_group)
        for plugin_name, plugin_entry in scanned_plugins:
            plugin_config = await self.etcd.get_prefix(
                f"config/plugins/{self._group_key}/{plugin_name}/"
            )
            try:
                plugin_instance = plugin_entry(plugin_config, self.local_config)
                await plugin_instance.init(context=context)
            except Exception:
                log.exception('error during initialization of plugin: {}', plugin_name)
                continue
            else:
                self.plugins[plugin_name] = plugin_instance
            if plugin_instance.config_watch_enabled:
                await self.watch_config_changes(plugin_name)
        await asyncio.sleep(0)

    async def cleanup(self) -> None:
        for wtask in {*self._config_watchers}:
            if not wtask.done():
                wtask.cancel()
                await wtask
        for plugin_instance in self.plugins.values():
            await plugin_instance.cleanup()

    async def _watcher(self, plugin_name: str) -> None:
        # As wait_timeout applies to the waiting for an internal async queue,
        # so short timeouts for polling the changes does not incur gRPC/network overheads.
        has_changes = False
        async for ev in self.etcd.watch_prefix(
            f"config/plugins/{self._group_key}/{plugin_name}",
            wait_timeout=0.2,
        ):
            if ev is QueueSentinel.TIMEOUT:
                if has_changes:
                    new_config = await self.etcd.get_prefix(
                        f"config/plugins/{self._group_key}/{plugin_name}/"
                    )
                    await self.plugins[plugin_name].update_plugin_config(new_config)
                has_changes = False
            else:
                has_changes = True

    async def watch_config_changes(self, plugin_name: str) -> None:
        wtask = asyncio.create_task(self._watcher(plugin_name))
        wtask.add_done_callback(self._config_watchers.discard)
        self._config_watchers.add(wtask)



import asyncio
import inspect
import time
from typing import (
    Any,
    Sequence,
    Dict,
)

import aioredis

__all__ = (
    'connect_with_retries',
    'execute_with_retries',
)

_scripts: Dict[str, str] = {}


def _calc_delay_exp_backoff(initial_delay: float, retry_count: float, time_limit: float) -> float:
    if time_limit > 0:
        return min(initial_delay * (2 ** retry_count), time_limit / 2)
    return min(initial_delay * (2 ** retry_count), 30.0)


async def connect_with_retries(
    *args,
    retry_delay: float = 0.5,
    retry_timeout: float = 60.0,
    max_retries: int = 0,
    exponential_backoff: bool = True,
    **kwargs,
) -> aioredis.abc.AbcPool:
    '''
    Create a Redis connection pool with multiple retries.
    '''
    begin = time.monotonic()
    num_retries = 0
    while True:
        try:
            return await aioredis.create_redis_pool(*args, **kwargs)
        except asyncio.CancelledError:
            raise
        except ConnectionRefusedError:
            if max_retries > 0 and num_retries >= max_retries:
                raise asyncio.TimeoutError('Exceeded the maximum retry count')
            if retry_timeout > 0 and time.monotonic() - begin >= retry_timeout:
                raise asyncio.TimeoutError('Too much delayed for retries')
            delay = _calc_delay_exp_backoff(
                retry_delay, num_retries, retry_timeout,
            ) if exponential_backoff else retry_delay
            await asyncio.sleep(delay)
            num_retries += 1
            continue


async def execute_with_retries(
    func: Any,
    retry_delay: float = 0.5,
    retry_timeout: float = 60.0,
    max_retries: int = 0,
    exponential_backoff: bool = True,
    suppress_force_closed: bool = True,
) -> Any:
    '''
    Execute the given Redis commands with multiple retries.
    The Redis commands must be generated as a ``aioredis.commands.Pipeline`` object
    or as a coroutine to execute single-shot aioredis commands by *func*.
    '''
    begin = time.monotonic()
    num_retries = 0
    while True:
        try:
            if inspect.iscoroutinefunction(func):
                aw_or_pipe = await func()
            elif callable(func):
                aw_or_pipe = func()
            else:
                raise TypeError('The func must be a function or a coroutinefunction '
                                'with no arguments.')
            if isinstance(aw_or_pipe, aioredis.commands.Pipeline):
                return await aw_or_pipe.execute()
            elif inspect.isawaitable(aw_or_pipe):
                return await aw_or_pipe
            else:
                raise TypeError('The return value must be an awaitable'
                                'or aioredis.commands.Pipeline object')
        except asyncio.CancelledError:
            raise
        except (aioredis.errors.PoolClosedError,
                aioredis.errors.ConnectionForcedCloseError):
            # This happens when we shut down the connection/pool.
            if suppress_force_closed:
                return None
            else:
                raise
        except (ConnectionResetError,
                ConnectionRefusedError,
                aioredis.errors.ConnectionClosedError,
                aioredis.errors.PipelineError):
            # Other cases mean server disconnection.
            if max_retries > 0 and num_retries >= max_retries:
                raise asyncio.TimeoutError('Exceeded the maximum retry count')
            if retry_timeout > 0 and time.monotonic() - begin >= retry_timeout:
                raise asyncio.TimeoutError('Too much delayed for retries')
            delay = _calc_delay_exp_backoff(
                retry_delay, num_retries, retry_timeout,
            ) if exponential_backoff else retry_delay
            await asyncio.sleep(delay)
            num_retries += 1
            continue


async def execute_script(
    conn: aioredis.Redis,
    script_id: str,
    script: str,
    keys: Sequence[str],
    args: Sequence[str],
) -> Any:
    """
    Auto-load and execute the given script.
    It uses the hash keys for scripts so that it does not send the whole
    script every time but only at the first time.

    Args:
        conn: A Redis connection or pool with the commands mixin.
        script_id: A human-readable identifier for the script.
            This can be arbitrary string but must be unique for each script.
        script: The script content.
        keys: The Redis keys that will be passed to the script.
        args: The arguments that will be passed to the script.
    """
    script_hash = _scripts.get(script_id, 'x')
    while True:
        try:
            ret = await execute_with_retries(lambda: conn.evalsha(
                script_hash,
                keys=keys,
                args=args,
            ))
            break
        except aioredis.errors.ReplyError as e:
            if 'NOSCRIPT' in e.args[0]:
                # Redis may have been restarted.
                script_hash = await conn.script_load(script)
                _scripts[script_id] = script_hash
            else:
                raise
            continue
    return ret



from __future__ import annotations

from abc import ABCMeta, abstractmethod
import enum
import logging
from typing import (
    Any,
    Final,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
)

import attr

from . import AbstractPlugin, BasePluginContext
from ..logging_utils import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__name__))

__all__ = (
    'HookHandler',
    'HookPlugin',
    'HookPluginContext',
    'Reject',
    'HookResults',
    'HookResult',
    'HookReturnTiming',
    'PASSED',
    'REJECTED',
    'ERROR',
    'ALL_COMPLETED',
    'FIRST_COMPLETED',
)


class HookHandler(Protocol):
    """
    The handler should accept a single argument containing
    a tuple of parameters passed to the handler.
    If it decides to cancel the ongoing event, it should raise
    :class:`HookDenied` exception.
    """

    async def __call__(self, *args) -> Any:
        # NOTE: Until https://github.com/python/mypy/issues/5876 is resolved,
        #       the get_handlers() in the HookPlugin subclasses should be marked
        #       with "type: ignore" comments.
        ...


class HookPlugin(AbstractPlugin, metaclass=ABCMeta):
    """
    The abstract interface for hook plugins.
    """

    @abstractmethod
    def get_handlers(self) -> Sequence[Tuple[str, HookHandler]]:
        """
        Returns a sequence of pairs of the event name
        and its corresponding handler function.
        """
        pass


class Reject(Exception):
    def __init__(self, reason: str):
        super().__init__(reason)
        self.reason = reason


class HookResults(enum.Enum):
    PASSED = 0
    REJECTED = 1
    ERROR = 2


class HookReturnTiming(enum.Enum):
    ALL_COMPLETED = 0
    FIRST_COMPLETED = 1


PASSED: Final = HookResults.PASSED
REJECTED: Final = HookResults.REJECTED
ERROR: Final = HookResults.ERROR
ALL_COMPLETED: Final = HookReturnTiming.ALL_COMPLETED
FIRST_COMPLETED: Final = HookReturnTiming.FIRST_COMPLETED


@attr.s(auto_attribs=True, slots=True, frozen=True)
class HookResult:
    status: HookResults
    src_plugin: Optional[Union[str, Sequence[str]]] = None
    reason: Optional[str] = None
    result: Optional[Any] = None


class HookPluginContext(BasePluginContext[HookPlugin]):
    """
    A manager for hook plugins with convenient handler invocation.
    """

    plugin_group = 'backendai_hook_v20'

    def _get_handlers(
        self, event_name: str, order: Sequence[str] = None,
    ) -> Sequence[Tuple[str, HookHandler]]:
        handlers = []
        for plugin_name, plugin_instance in self.plugins.items():
            for hooked_event_name, hook_handler in plugin_instance.get_handlers():
                if event_name != hooked_event_name:
                    continue
                handlers.append((plugin_name, hook_handler))
        if order is not None:
            non_empty_order = order
            handlers.sort(key=lambda item: non_empty_order.index(item))
        else:
            # the default is alphabetical order with plugin names
            handlers.sort(key=lambda item: item[0])
        return handlers

    async def dispatch(
        self, event_name: str, args: Tuple[Any, ...], *,
        return_when: HookReturnTiming = ALL_COMPLETED,
        success_if_no_hook: bool = True,
        order: Sequence[str] = None,
    ) -> HookResult:
        """
        Invoke the handlers that matches with the given ``event_name``.
        If any of the handlers raises :class:`HookDenied`,
        the event caller should seize the processing.
        """
        executed_plugin_names = []
        results: List[Any] = []
        for plugin_name, hook_handler in self._get_handlers(event_name, order=order):
            try:
                executed_plugin_names.append(plugin_name)
                result = await hook_handler(*args)
            except Reject as e:
                return HookResult(
                    status=REJECTED,
                    src_plugin=plugin_name,
                    reason=e.reason,
                )
            except Exception as e:
                return HookResult(
                    status=ERROR,
                    src_plugin=plugin_name,
                    reason=repr(e),
                )
            else:
                if return_when == FIRST_COMPLETED:
                    return HookResult(
                        status=PASSED,
                        src_plugin=plugin_name,
                        result=result,
                    )
                else:
                    results.append(result)
        if not success_if_no_hook and not executed_plugin_names:
            return HookResult(
                status=REJECTED,
                src_plugin=executed_plugin_names,  # empty
                result=results,                    # empty
            )
        return HookResult(
            status=PASSED,
            src_plugin=executed_plugin_names,
            result=results,
        )

    async def notify(
        self, event_name: str, args: Tuple[Any, ...],
    ) -> None:
        """
        Invoke the handlers that matches with the given ``event_name``.
        Regardless of the handler results, the processing continues.
        """
        for plugin_name, hook_handler in self._get_handlers(event_name):
            try:
                await hook_handler(*args)
            except Exception:
                log.exception('HookPluginContext.notify({}): skipping error in hook handler from {}',
                              event_name, plugin_name)
                continue
